# -*- coding: utf-8 -*-
import os
import re
import socket
from stat import S_ISDIR
from datetime import datetime
import logging
import paramiko


MTIME_TOLERANCE = 3
logger = logging.getLogger(__name__)


class AuthenticationError(Exception):
    pass


class TimeoutError(Exception):
    pass


class SshError(Exception):
    pass


class Sftp(object):

    def __init__(self, host, username, password=None, port=22, timeout=10,
                 max_attempts=3, **kwargs):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.client = paramiko.SSHClient()
        self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.sftp = None
        for i in range(max_attempts):
            try:
                self.client.connect(host, port=port, username=username,
                                    password=password, timeout=timeout,
                                    **kwargs)
                self.sftp = self.client.open_sftp()
                return
            except (paramiko.BadHostKeyException,
                    paramiko.AuthenticationException), e:
                raise AuthenticationError(str(e))
            except socket.timeout, e:
                raise TimeoutError(str(e))
            except Exception, e:
                if i == max_attempts - 1:
                    raise SshError(str(e))

    def _walk_remote(self, path, topdown=True):
        try:
            res = self.sftp.listdir_attr(path)
        except IOError:
            res = []

        for stat in res:
            file = os.path.join(path, stat.filename)

            if not S_ISDIR(stat.st_mode):
                yield 'file', file, stat
            else:
                if topdown:
                    yield 'dir', file, stat
                    for res in self._walk_remote(file, topdown=topdown):
                        yield res
                else:
                    for res in self._walk_remote(file, topdown=topdown):
                        yield res
                    yield 'dir', file, None

    def _walk_local(self, path, topdown=True):
        for path, dirs, files in os.walk(path, topdown=topdown):
            for file in files:
                file = os.path.join(path, file)
                yield 'file', file, os.stat(file)
            for dir in dirs:
                dir = os.path.join(path, dir)
                yield 'dir', dir, os.stat(dir)

    def _walk(self, *args, **kwargs):
        remote = kwargs.pop('remote', False)
        if remote:
            return self._walk_remote(*args, **kwargs)
        else:
            return self._walk_local(*args, **kwargs)

    def _makedirs_dst(self, path, remote=True, dry=False):
        if remote:
            paths = []
            while path not in ('/', ''):
                paths.insert(0, path)
                path = os.path.dirname(path)

            for path in paths:
                try:
                    self.sftp.lstat(path)
                except Exception:
                    if not dry:
                        self.sftp.mkdir(path)
                    logger.debug('created destination directory %s', path)
        else:
            if not os.path.exists(path):
                if not dry:
                    os.makedirs(path)
                logger.debug('created destination directory %s', path)

    def _validate_src(self, file, include, exclude):
        for re_ in include:
            if not re_.search(file):
                return False
        for re_ in exclude:
            if re_.search(file):
                return False
        return True

    def _validate_dst(self, file, src_stat, remote=True):
        if remote:
            try:
                dst_stat = self.sftp.lstat(file)
            except Exception:
                return
        else:
            if not os.path.exists(file):
                return
            dst_stat = os.stat(file)

        if abs(dst_stat.st_mtime - src_stat.st_mtime) > MTIME_TOLERANCE:
            debugString = '%s modified time mismatch '
            debugString += '(source: %s, destination: %s)'
            logger.debug(debugString,
                         file, datetime.utcfromtimestamp(src_stat.st_mtime),
                         datetime.utcfromtimestamp(dst_stat.st_mtime))
            return
        if dst_stat.st_size != src_stat.st_size:
            return
        return True

    def _save(self, src, dst, src_stat, remote=True):
        if remote:
            logger.info('copying %s to %s@%s:%s', src, self.username,
                        self.host, dst)
            self.sftp.put(src, dst)
            self.sftp.utime(dst, (int(src_stat.st_atime),
                            int(src_stat.st_mtime)))
        else:
            logger.info('copying %s@%s:%s to %s', self.username, self.host,
                        src, dst)
            self.sftp.get(src, dst)
            os.utime(dst, (int(src_stat.st_atime), int(src_stat.st_mtime)))

    def _delete_dst(self, path, files, remote=True, dry=False):
        if remote:
            callables = {'file': self.sftp.remove, 'dir': self.sftp.rmdir}
        else:
            callables = {'file': os.remove, 'dir': os.rmdir}

        for type, file, stat in self._walk(path, topdown=False, remote=remote):
            if file not in files[type]:
                if not dry:
                    try:
                        callables[type](file)
                    except Exception, e:
                        logger.debug('failed to remove %s: %s', file, str(e))
                        continue

                logger.debug('removed %s', file)

    def _get_filters(self, filters):
        if not filters:
            return []
        return [re.compile(f) for f in filters]

    def sync(self, src, dst, download=True, include=None, exclude=None,
             delete=False, dry=False):
        '''Sync files and directories.

        :param src: source directory
        :param dst: destination directory
        :param download: True to sync from a remote source to a local
            destination, else sync from a local source to a remote destination
        :param include: list of regex patterns the source files must match
        :param exclude: list of regex patterns the source files must not match
        :param delete: remove destination files and directories not present
            at source or filtered by the include/exlude patterns
        '''
        include = self._get_filters(include)
        exclude = self._get_filters(exclude)

        if src.endswith('/') != dst.endswith('/'):
            dst = os.path.join(dst, os.path.basename(src.rstrip('/')))
        src = src.rstrip('/')
        re_base = re.compile(r'^%s/' % re.escape(src))
        if not src:
            src = '/'

        self._makedirs_dst(dst, remote=not download, dry=dry)

        started = datetime.utcnow()
        total_size = 0
        dst_list = {'file': [], 'dir': []}

        for type, file, stat in self._walk(src, remote=download):
            file_ = re_base.sub('', file)
            if not self._validate_src(file_, include, exclude):
                logger.debug('filtered %s', file)
                continue

            dst_file = os.path.join(dst, file_)
            dst_list[type].append(dst_file)

            if type == 'dir':
                self._makedirs_dst(dst_file, remote=not download, dry=dry)
            elif type == 'file':
                if not self._validate_dst(dst_file, stat, remote=not download):
                    if not dry:
                        self._save(file, dst_file, stat, remote=not download)
                    total_size += stat.st_size
                    logger.debug('copied %s to %s', file, dst_file)

        if delete:
            self._delete_dst(dst, dst_list, remote=not download, dry=dry)

        logger.debug('transferred %s bytes in %s', total_size,
                     datetime.utcnow() - started)



# -*- coding: utf-8 -*-

from jinja2 import Environment, FileSystemLoader
import markdown2
import os
import shutil
import subprocess
import yaml
import json


def renderSiteOuput(sourceDir, outputBase, templatesDir, siteData):
    for root, dirs, files in os.walk(sourceDir):
        for f in files:
            folderID = root[len(sourceDir):]
            outputDir = outputBase + folderID
            if f[:1] != '.':
                renderSiteFile(root, f, outputDir, templatesDir, siteData,
                               folderID)


def populateSiteData(location):
    siteData = {'/': []}
    for root, dirs, files in os.walk(location):
        folderID = '/' + root[len(location) + 1:]
        siteData[folderID] = []
        for f in files:
            if os.path.splitext(f)[1] == '.md':
                mdData = dictFromMarkdown(os.path.join(root, f), folderID)
                siteData[folderID].append(mdData)
            if os.path.splitext(f)[1] == '.yaml':
                ymlData = dictFromYaml(os.path.join(root, f), folderID)
                siteData[folderID].append(ymlData)
        for folder in dirs:
            yamlData = {'relativePath': folderID}
            yamlData['type'] = 'folder'
            yamlData['name'] = folder
            siteData[folderID].append(yamlData)
    return siteData


def dictFromMarkdown(markdownPath, folderID):
    separator = '-*-*-*-\n'
    rawFile = open(markdownPath).read()
    separatorIndex = rawFile.find(separator)
    if separatorIndex > -1:
        yamlData = yaml.load(rawFile[:separatorIndex])
        mdData = rawFile[separatorIndex + len(separator):]
        htmlData = markdown2.markdown(mdData)
        yamlData['html'] = htmlData
        yamlData['type'] = 'yaml+markdown'
    else:
        yamlData = {'html': markdown2.markdown(rawFile)}
        yamlData['type'] = 'markdown'
    yamlData['relativePath'] = folderID
    yamlData['url'] = folderID + '/'
    filename = os.path.basename(markdownPath)
    yamlData['url'] += os.path.splitext(filename)[0] + '.html'
    return yamlData


def dictFromYaml(yamlPath, folderID):
    yamlData = yaml.load(open(yamlPath).read())
    yamlData['relativePath'] = folderID
    yamlData['type'] = 'yaml'
    if 'template' in yamlData:
        yamlData['url'] = folderID + '/'
        filename = os.path.basename(yamlPath)
        yamlData['url'] += os.path.splitext(filename)[0] + '.html'
    return yamlData


def renderMarkdownFile(sourceDir, filename, newDir, templatesDir, siteData,
                       folderID):
    newFilename = os.path.splitext(filename)[0] + '.html'
    newPath = os.path.join(newDir, newFilename)
    mdData = dictFromMarkdown(os.path.join(sourceDir, filename), folderID)
    if 'template' in mdData:
        jinjaEnv = Environment(loader=FileSystemLoader([templatesDir]))
        template = jinjaEnv.get_template(mdData['template'])
        output = template.render(page=mdData, site=siteData)
        with open(newPath, 'w') as f:
            f.write(output.encode('utf-8'))
    else:
        with open(newPath, 'w') as f:
            f.write(mdData['html'].encode('utf-8'))
    return newPath


def renderYamlFile(sourceDir, filename, newDir, templatesDir, siteData,
                   folderID):
    newFilename = os.path.splitext(filename)[0] + '.html'
    newPath = os.path.join(newDir, newFilename)
    yamlData = dictFromYaml(os.path.join(sourceDir, filename), folderID)
    if 'template' in yamlData:
        jinjaEnv = Environment(loader=FileSystemLoader([templatesDir]))
        template = jinjaEnv.get_template(yamlData['template'])
        output = template.render(page=yamlData, site=siteData)
        with open(newPath, 'w') as f:
            f.write(output.encode('utf-8'))
    else:
        return False
    return newPath


def renderHtmlFile(sourceDir, filename, newDir, templatesDir, siteData,
                   folderID):
    newPath = os.path.join(newDir, filename)
    jinjaFLS = FileSystemLoader([templatesDir, sourceDir])
    jinjaEnv = Environment(loader=jinjaFLS)
    template = jinjaEnv.get_template(filename)
    output = template.render(site=siteData)
    with open(newPath, 'w') as f:
        f.write(output.encode('utf-8'))
    return newPath


def renderSiteFile(sourceDir, filename, outputDir, templatesDir, siteData,
                   folderID):
    extension = os.path.splitext(filename)[1]
    currentPath = os.path.join(sourceDir, filename)
    source_length = len(sourceDir) + 1
    newDir = os.path.join(outputDir, sourceDir[source_length:])
    if not os.path.exists(newDir):
        os.makedirs(newDir)
    if extension == '.html':
        newPath = renderHtmlFile(sourceDir, filename, newDir,
                                 templatesDir, siteData, folderID)
    elif extension == '.md':
        newPath = renderMarkdownFile(sourceDir, filename, newDir,
                                     templatesDir, siteData, folderID)
    elif extension == '.yaml':
        newPath = renderYamlFile(sourceDir, filename, newDir,
                                 templatesDir, siteData, folderID)
    else:
        newPath = os.path.join(newDir, filename)
        shutil.copy2(currentPath, newDir)
    return newPath


def renderLessFile(lessFile, outputDir):
    basename = os.path.basename(lessFile)
    cssDir = os.path.join(outputDir, 'css')
    if not os.path.exists(cssDir):
        os.makedirs(cssDir)
    cssFile = os.path.splitext(basename)[0] + '.css'
    outputFile = os.path.join(cssDir, cssFile)
    subCall = ['lessc', '-x', lessFile, outputFile]
    subprocess.call(subCall)
    return outputFile


def renderSassFile(sassFile, outputDir):
    basename = os.path.basename(sassFile)
    cssDir = os.path.join(outputDir, 'css')
    if not os.path.exists(cssDir):
        os.makedirs(cssDir)
    cssFile = os.path.splitext(basename)[0] + '.css'
    outputFile = os.path.join(cssDir, cssFile)
    subCall = ['sass', '--style', 'compressed', sassFile, outputFile]
    subprocess.call(subCall)
    return outputFile


def deleteFolderContents(folder_path):
    for file_object in os.listdir(folder_path):
        file_object_path = os.path.join(folder_path, file_object)
        if os.path.isfile(file_object_path):
            os.unlink(file_object_path)
        else:
            shutil.rmtree(file_object_path)


def buildStaticSite(sourceDir, outputDir, templatesDir=False, lessFile=False,
                    sassFile=False, printSiteData=False):
    deleteFolderContents(outputDir)
    siteData = populateSiteData(sourceDir)
    if printSiteData:
        print json.dumps(siteData, sort_keys=True, indent=4)
    if not templatesDir:
        templatesDir = os.path.join(os.path.dirname(sourceDir), 'templates')
    renderSiteOuput(sourceDir, outputDir, templatesDir, siteData)
    if lessFile:
        renderLessFile(lessFile, outputDir)
    if sassFile:
        renderSassFile(sassFile, outputDir)



# -*- coding: utf-8 -*-

import os


def createNewSite(location, cssType):
    print 'creating new jagss site...'
    siteName = os.path.split(location)[1]
    sitePath = os.path.join(location, 'site')
    outputPath = os.path.join(location, 'staticOutput')
    templatesPath = os.path.join(location, 'templates')
    configPath = os.path.join(location, 'config.yaml')
    dirsToCreate = [sitePath, outputPath, templatesPath]
    jinjaPath = os.path.join(templatesPath, 'default.html')
    indexPath = os.path.join(sitePath, 'index.md')
    if cssType == 'less':
        lessDirPath = os.path.join(location, 'less')
        cssFilePath = os.path.join(lessDirPath, siteName + '.less')
        dirsToCreate.append(lessDirPath)
    elif cssType == 'sass':
        sassDirPath = os.path.join(location, 'sass')
        cssFilePath = os.path.join(sassDirPath, siteName + '.scss')
        dirsToCreate.append(sassDirPath)
    else:
        cssDirPath = os.path.join(sitePath, 'css')
        cssFilePath = os.path.join(cssDirPath, siteName + '.css')
        dirsToCreate.append(cssDirPath)
    for dirPath in dirsToCreate:
        if not os.path.exists(dirPath):
            os.makedirs(dirPath)
    if not os.path.exists(configPath):
        createConfigFile(location, cssType)
    if not os.path.exists(cssFilePath):
        createCssFile(cssFilePath)
    if not os.path.exists(jinjaPath):
        createTemplateFile(jinjaPath, siteName)
    if not os.path.exists(indexPath):
        createMarkdownFile(indexPath)


def createConfigFile(cwd, cssType):
    """creates a config.yaml file with default values inside the cwd"""
    siteName = os.path.split(cwd)[1]
    conf = 'sourceDir: ' + os.path.join(cwd, 'site') + '\n'
    conf += 'outputDir: ' + os.path.join(cwd, 'staticOutput') + '\n'
    conf += 'templatesDir: ' + os.path.join(cwd, 'templates') + '\n'
    if cssType == 'less':
        lessPath = os.path.join(cwd, 'less')
        conf += 'lessFile: ' + os.path.join(lessPath, siteName + '.less')
        conf += '\n'
        conf += 'sassFile: False\n'
    elif cssType == 'sass':
        conf += 'lessFile: False\n'
        sassPath = os.path.join(cwd, 'sass')
        conf += 'sassFile: ' + os.path.join(sassPath, siteName + '.scss')
        conf += '\n'
    else:
        conf += 'lessFile: False\n'
        conf += 'sassFile: False\n'
    conf += '\n# deployment credentials go here:\n'
    conf += 'sftpAddress: none\n'
    conf += 'sftpUsername: none\n'
    conf += 'sftpPassword: none\n'
    conf += 'sftpDirectory: none\n'
    conf += 's3AccessKey: none\n'
    conf += 's3SecretKey: none\n'
    conf += 's3BucketName: none\n'
    with open(os.path.join(cwd, 'config.yaml'), 'w') as f:
        f.write(conf.encode('utf-8'))


def createCssFile(cssPath):
    """creates a default css or less file at the specified path."""
    css = """body {
    width: 600px;
    margin: 150px auto;
    font-family: sans-serif;
    background-color: #008209;
    color: #eee;
}
"""
    with open(cssPath, 'w') as f:
        f.write(css.encode('utf-8'))


def createTemplateFile(templatePath, siteName):
    """creates a default jinja2 template at the specified path."""
    template = """<!DOCTYPE html>
<html lang="en">
    <head>
        <meta charset="utf-8">
        <meta name="viewport" content="width=device-width, initial-scale=1.0">
        <title>{{ page.title }}</title>
        <link href="/css/%s.css" rel="stylesheet" />
    </head>
    <body>
        {{ page.html }}
    </body>
</html>
""" % (siteName)
    with open(templatePath, 'w') as f:
        f.write(template.encode('utf-8'))


def createMarkdownFile(markdownPath):
    """creates a default markdown file at the specified path."""
    md = """template: default.html
title: JAGSS!
-*-*-*-
# Welcome to JAGSS!

This is the default page for a new JAGSS project. Modify the template
and site files to get started.
"""
    with open(markdownPath, 'w') as f:
        f.write(md.encode('utf-8'))



from setuptools import setup, find_packages


def readme():
    with open('readme.md') as f:
        return f.read()

requirements = [
    'jinja2>=2.0',
    'markdown2>=2.0',
    'pyyaml>=3.10',
    'watchdog>=0.6',
    'paramiko>=1.12.0',
    'boto_rsync>=0.8.1'
]


setup(name='jagss',
      version='0.0.2',
      description='Just another generator for static sites',
      url='https://jagss.rpy.xyz',
      author='Evan Sonderegger',
      author_email='evan@rpy.xyz',
      license='MIT',
      packages=find_packages(exclude=['examples*']),
      install_requires=requirements,
      scripts=['bin/jagss'],
      test_suite='nose.collector',
      tests_require=['nose'],
      zip_safe=False)
